import { ISequence } from '../algorithm/sequence';
import { Message } from '../core/messaging';
import { ISignal } from '../core/signaling';
import { TabPanel } from './tabpanel';
import { Widget } from './widget';
/**
 * A widget which provides a flexible docking area for widgets.
 */
export declare class DockPanel extends Widget {
    /**
     * Construct a new dock panel.
     *
     * @param options - The options for initializing the panel.
     */
    constructor(options?: DockPanel.IOptions);
    /**
     * Dispose of the resources held by the panel.
     */
    dispose(): void;
    /**
     * A signal emitted when the current widget has changed.
     */
    currentChanged: ISignal<DockPanel, DockPanel.ICurrentChangedArgs>;
    /**
     * The overlay used by the dock panel.
     *
     * #### Notes
     * This is a read-only property.
     */
    overlay: DockPanel.IOverlay;
    /**
     * Get the spacing between the panels.
     */
    /**
     * Set the spacing between the panels.
     */
    spacing: number;
    /**
     * A read-only sequence of the widgets in the dock panel.
     *
     * #### Notes
     * The order of the widgets in the sequence has no meaning.
     *
     * This is a read-only property.
     */
    widgets: ISequence<Widget>;
    /**
     * The current widget in the dock panel.
     *
     * #### Notes
     * The current widget is the widget among the added widgets which
     * has the *descendant node* which has most recently been focused.
     *
     * This is the `currentWidget` of the internal `FocusTracker` which
     * tracks all widgets in the dock panel.
     *
     * This will be `null` if there is no current widget.
     *
     * This is a read-only property.
     */
    currentWidget: Widget;
    /**
     * Activate the specified widget in the dock panel.
     *
     * @param widget - The widget of interest.
     *
     * #### Notes
     * This will ensure that the widget is the current visible widget
     * in its host tab panel and post the widget an activate request.
     */
    activateWidget(widget: Widget): void;
    /**
     * Add a widget to the dock panel.
     *
     * @param widget - The widget to add into the dock panel.
     *
     * @param options - The additional options for adding the widget.
     */
    addWidget(widget: Widget, options?: DockPanel.IAddOptions): void;
    /**
     * Find the drop target for the given client position.
     *
     * @param clientX - The client X position of interest.
     *
     * @param clientY - The client Y position of interest.
     *
     * @returns The drop target at the specified client position.
     */
    findDropTarget(clientX: number, clientY: number): DockPanel.IDropTarget;
    /**
     * Show the overlay indicator at the given client position.
     *
     * @param clientX - The client X position of interest.
     *
     * @param clientY - The client Y position of interest.
     *
     * @returns The dock zone at the specified client position.
     *
     * #### Notes
     * If the position is not over a valid zone, the overlay is hidden.
     *
     * This returns the dock zone used to display the overlay.
     */
    showOverlay(clientX: number, clientY: number): DockPanel.Zone;
    /**
     * Handle the DOM events for the dock panel.
     *
     * @param event - The DOM event sent to the dock panel.
     *
     * #### Notes
     * This method implements the DOM `EventListener` interface and is
     * called in response to events on the dock panel's node. It should
     * not be called directly by user code.
     */
    handleEvent(event: Event): void;
    /**
     * A message handler invoked on an `'after-attach'` message.
     */
    protected onAfterAttach(msg: Message): void;
    /**
     * A message handler invoked on a `'before-detach'` message.
     */
    protected onBeforeDetach(msg: Message): void;
    /**
     * Handle the `'p-dragenter'` event for the dock panel.
     */
    private _evtDragEnter(event);
    /**
     * Handle the `'p-dragleave'` event for the dock panel.
     */
    private _evtDragLeave(event);
    /**
     * Handle the `'p-dragover'` event for the dock panel.
     */
    private _evtDragOver(event);
    /**
     * Handle the `'p-drop'` event for the dock panel.
     */
    private _evtDrop(event);
    /**
     * Drop a widget onto the dock panel using the given drop target.
     */
    private _handleDrop(widget, target);
    /**
     * Insert a widget into the dock panel using the given mode.
     *
     * The target widget should have no parent, and the reference widget
     * should either be null or a widget contained in the dock panel.
     */
    private _insertWidget(widget, mode, ref);
    /**
     * Create a new tab panel for adding to the dock panel.
     */
    private _createTabPanel();
    /**
     * Create a new split panel for adding to the dock panel.
     */
    private _createSplitPanel(orientation);
    /**
     * Dispose of a tab panel created for the dock panel.
     */
    private _disposeTabPanel(panel);
    /**
     * Dispose of a split panel created for the dock panel.
     */
    private _disposeSplitPanel(panel);
    /**
     * Ensure a tab panel is available in the dock panel.
     *
     * If there is no root panel, a new root tab panel will be created.
     * Otherwise, if there is an active widget, the tab panel for that
     * widget will be returned. Otherwise, the top-left tab panel will
     * be returned.
     */
    private _ensureTabPanel();
    /**
     * Ensure the root panel is a split panel with the given orientation.
     *
     * This returns split panel so that casting the root is not needed.
     */
    private _ensureSplitRoot(orientation);
    /**
     * Set the non-null root widget for the dock panel.
     */
    private _setRoot(root);
    /**
     * Remove an empty dock tab panel from the hierarchy.
     *
     * This ensures that the hierarchy is kept consistent by merging an
     * ancestor split panel when it contains only a single child widget.
     */
    private _removeTabPanel(tabPanel);
    /**
     * Handle the `tabDetachRequested` signal from a tab bar.
     */
    private _onTabDetachRequested(sender, args);
    /**
     * Handle the `widgetRemoved` signal from a stacked panel.
     */
    private _onWidgetRemoved(sender, widget);
    /**
     * Handle the `currentChanged` signal from the focus tracker.
     */
    private _onCurrentChanged(sender, args);
    private _spacing;
    private _drag;
    private _overlay;
    private _widgets;
    private _tabPanels;
    private _splitPanels;
    private _tracker;
    private _root;
}
/**
 * The namespace for the `DockPanel` class statics.
 */
export declare namespace DockPanel {
    /**
     * An arguments object for the `currentChanged` signal.
     */
    interface ICurrentChangedArgs {
        /**
         * The old value for the `currentWidget`, or `null`.
         */
        oldValue: Widget;
        /**
         * The new value for the `currentWidget`, or `null`.
         */
        newValue: Widget;
    }
    /**
     * An options object for creating a dock panel.
     */
    interface IOptions {
        /**
         * The spacing between tab panels in the dock panel.
         *
         * The default is `4`.
         */
        spacing?: number;
        /**
         * The overlay to use with the dock panel.
         *
         * The default is a new `Overlay` instance.
         */
        overlay?: IOverlay;
    }
    /**
     * An options object for adding a widget to the dock panel.
     */
    interface IAddOptions {
        /**
         * The insertion mode for adding the widget.
         *
         * The default is `'tab-after'`.
         */
        mode?: Mode;
        /**
         * The reference widget for the insert location.
         *
         * The default is `null`.
         */
        ref?: Widget;
        /**
         * Whether to activate the new widget.
         *
         * The default is `true`.
         */
        activate?: boolean;
    }
    /**
     * A type alias for the supported insertion modes.
     *
     * A dock mode is used to specify how a widget should be added
     * to the dock panel relative to a reference widget.
     */
    type Mode = ('split-top' | 'split-left' | 'split-right' | 'split-bottom' | 'tab-before' | 'tab-after');
    /**
     * A type alias of the supported dock zones.
     *
     * A dock zone is used to indicate the position of the dock panel
     * overlay when the user drags over the dock panel with the mouse.
     */
    type Zone = ('invalid' | 'root-top' | 'root-left' | 'root-right' | 'root-bottom' | 'root-center' | 'panel-top' | 'panel-left' | 'panel-right' | 'panel-bottom' | 'panel-center');
    /**
     * The result object for a drop target hit test.
     */
    interface IDropTarget {
        /**
         * The dock zone at the specified position.
         *
         * This will be `'invalid'` if the position is not a valid zone.
         */
        zone: Zone;
        /**
         * The tab panel for the indicated dock zone.
         *
         * This will be `null` if the dock zone is not a panel zone.
         */
        panel: TabPanel;
    }
    /**
     * An object which holds the geometry for overlay positioning.
     */
    interface IOverlayGeometry {
        /**
         * The zone associated with the geometry.
         */
        zone: Zone;
        /**
         * The client X position of the mouse.
         */
        mouseX: number;
        /**
         * The client Y position of the mouse.
         */
        mouseY: number;
        /**
         * The client rect of the overlay parent node.
         */
        parentRect: ClientRect;
        /**
         * The distance between the overlay and parent top edges.
         */
        top: number;
        /**
         * The distance between the overlay and parent left edges.
         */
        left: number;
        /**
         * The distance between the overlay and parent right edges.
         */
        right: number;
        /**
         * The distance between the overlay and parent bottom edges.
         */
        bottom: number;
        /**
         * The width of the overlay.
         */
        width: number;
        /**
         * The height of the overlay.
         */
        height: number;
    }
    /**
     * An object which manages the overlay node for a dock panel.
     */
    interface IOverlay {
        /**
         * The DOM node for the overlay.
         *
         * #### Notes
         * This should be a read-only property.
         */
        node: HTMLDivElement;
        /**
         * Show the overlay using the given overlay geometry.
         *
         * @param geo - The desired geometry for the overlay.
         *
         * #### Notes
         * The given geometry values assume the node will use absolute
         * positioning.
         *
         * This is called on every mouse move event during a drag in order
         * to update the position of the overlay. It should be efficient.
         */
        show(geo: IOverlayGeometry): void;
        /**
         * Hide the overlay node.
         *
         * @param delay - The delay (in ms) before hiding the overlay.
         *   A delay value <= 0 should hide the overlay immediately.
         *
         * #### Notes
         * This is called whenever the overlay node should been hidden.
         */
        hide(delay: number): void;
    }
    /**
     * A concrete implementation of `IOverlay`.
     *
     * This is the default overlay implementation for a dock panel.
     */
    class Overlay implements IOverlay {
        /**
         * Construct a new overlay.
         */
        constructor();
        /**
         * The DOM node for the overlay.
         *
         * #### Notes
         * This is a read-only property.
         */
        node: HTMLDivElement;
        /**
         * Show the overlay using the given overlay geometry.
         *
         * @param geo - The desired geometry for the overlay.
         */
        show(geo: IOverlayGeometry): void;
        /**
         * Hide the overlay node.
         *
         * @param delay - The delay (in ms) before hiding the overlay.
         *   A delay value <= 0 will hide the overlay immediately.
         */
        hide(delay: number): void;
        /**
         * Set the dock zone for the overlay.
         */
        private _setZone(newZone);
        private _timer;
        private _hidden;
        private _node;
        private _zone;
    }
}
