"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
/*-----------------------------------------------------------------------------
| Copyright (c) 2014-2016, PhosphorJS Contributors
|
| Distributed under the terms of the BSD 3-Clause License.
|
| The full license is in the file LICENSE, distributed with this software.
|----------------------------------------------------------------------------*/
var iteration_1 = require('../algorithm/iteration');
var searching_1 = require('../algorithm/searching');
var vector_1 = require('../collections/vector');
var mimedata_1 = require('../core/mimedata');
var signaling_1 = require('../core/signaling');
var dragdrop_1 = require('../dom/dragdrop');
var query_1 = require('../dom/query');
var sizing_1 = require('../dom/sizing');
var focustracker_1 = require('./focustracker');
var splitpanel_1 = require('./splitpanel');
var stackedpanel_1 = require('./stackedpanel');
var tabpanel_1 = require('./tabpanel');
var widget_1 = require('./widget');
/**
 * The class name added to a DockPanel instance.
 */
var DOCK_PANEL_CLASS = 'p-DockPanel';
/**
 * The class name added to a DockPanel tab panel descendant.
 */
var TAB_PANEL_CLASS = 'p-DockPanel-tabPanel';
/**
 * The class name added to a DockPanel split panel descendant.
 */
var SPLIT_PANEL_CLASS = 'p-DockPanel-splitPanel';
/**
 * The class name added to a DockPanel overlay.
 */
var OVERLAY_CLASS = 'p-DockPanel-overlay';
/**
 * The class name added to hidden entities.
 */
var HIDDEN_CLASS = 'p-mod-hidden';
/**
 * The factory MIME type supported by the dock panel.
 */
var FACTORY_MIME = 'application/vnd.phosphor.widget-factory';
/**
 * The size of the edge dock zone for the root panel, in pixels.
 */
var EDGE_SIZE = 30;
/**
 * A widget which provides a flexible docking area for widgets.
 */
var DockPanel = (function (_super) {
    __extends(DockPanel, _super);
    /**
     * Construct a new dock panel.
     *
     * @param options - The options for initializing the panel.
     */
    function DockPanel(options) {
        if (options === void 0) { options = {}; }
        _super.call(this);
        this._drag = null;
        this._widgets = new vector_1.Vector();
        this._tabPanels = new vector_1.Vector();
        this._splitPanels = new vector_1.Vector();
        this._tracker = new focustracker_1.FocusTracker();
        this._root = null;
        this.addClass(DOCK_PANEL_CLASS);
        // Install the layout on the panel.
        this.layout = new stackedpanel_1.StackedLayout();
        // Parse the spacing option.
        if (options.spacing !== void 0) {
            this._spacing = Private.clampSpacing(options.spacing);
        }
        // Setup the overlay indicator.
        if (options.overlay !== void 0) {
            this._overlay = options.overlay;
        }
        else {
            this._overlay = new DockPanel.Overlay();
        }
        // Connect the focus tracker changed signal.
        this._tracker.currentChanged.connect(this._onCurrentChanged, this);
        // Add the overlay node to the panel.
        this.node.appendChild(this._overlay.node);
    }
    /**
     * Dispose of the resources held by the panel.
     */
    DockPanel.prototype.dispose = function () {
        // Hide the overlay.
        this._overlay.hide(0);
        // Cancel a drag if one is in progress.
        if (this._drag)
            this._drag.dispose();
        // Clear the data structures.
        this._root = null;
        this._widgets.clear();
        this._tabPanels.clear();
        this._splitPanels.clear();
        // Dispose of the focus tracker.
        this._tracker.dispose();
        // Dispose of the base class.
        _super.prototype.dispose.call(this);
    };
    Object.defineProperty(DockPanel.prototype, "overlay", {
        /**
         * The overlay used by the dock panel.
         *
         * #### Notes
         * This is a read-only property.
         */
        get: function () {
            return this._overlay;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DockPanel.prototype, "spacing", {
        /**
         * Get the spacing between the panels.
         */
        get: function () {
            return this._spacing;
        },
        /**
         * Set the spacing between the panels.
         */
        set: function (value) {
            value = Private.clampSpacing(value);
            if (this._spacing === value) {
                return;
            }
            this._spacing = value;
            iteration_1.each(this._splitPanels, function (panel) { panel.spacing = value; });
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DockPanel.prototype, "widgets", {
        /**
         * A read-only sequence of the widgets in the dock panel.
         *
         * #### Notes
         * The order of the widgets in the sequence has no meaning.
         *
         * This is a read-only property.
         */
        get: function () {
            return this._widgets;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DockPanel.prototype, "currentWidget", {
        /**
         * The current widget in the dock panel.
         *
         * #### Notes
         * The current widget is the widget among the added widgets which
         * has the *descendant node* which has most recently been focused.
         *
         * This is the `currentWidget` of the internal `FocusTracker` which
         * tracks all widgets in the dock panel.
         *
         * This will be `null` if there is no current widget.
         *
         * This is a read-only property.
         */
        get: function () {
            return this._tracker.currentWidget;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Activate the specified widget in the dock panel.
     *
     * @param widget - The widget of interest.
     *
     * #### Notes
     * This will ensure that the widget is the current visible widget
     * in its host tab panel and post the widget an activate request.
     */
    DockPanel.prototype.activateWidget = function (widget) {
        // Ensure the widget is contained by the panel.
        if (searching_1.indexOf(this._widgets, widget) === -1) {
            throw new Error('Widget is not contained by the dock panel.');
        }
        // Ensure the widget is the current widget.
        widget.parent.parent.currentWidget = widget;
        // Activate the widget.
        widget.activate();
    };
    /**
     * Add a widget to the dock panel.
     *
     * @param widget - The widget to add into the dock panel.
     *
     * @param options - The additional options for adding the widget.
     */
    DockPanel.prototype.addWidget = function (widget, options) {
        if (options === void 0) { options = {}; }
        // Setup the option defaults.
        var activate = true;
        var ref = null;
        var mode = 'tab-after';
        // Extract the options.
        if (options.activate !== void 0) {
            activate = options.activate;
        }
        if (options.ref !== void 0) {
            ref = options.ref;
        }
        if (options.mode !== void 0) {
            mode = options.mode;
        }
        // Ensure the arguments are valid.
        if (!widget) {
            throw new Error('Target widget is null.');
        }
        if (widget === ref) {
            throw new Error('Target widget cannot be the reference widget.');
        }
        if (ref && searching_1.indexOf(this._widgets, ref) === -1) {
            throw new Error('Reference widget is not contained by the dock panel.');
        }
        // Unparent the widget before performing the insert, so that structural
        // changes to the panel occur before searching for the insert location.
        widget.parent = null;
        // Add the widget to the focus tracker.
        this._tracker.add(widget);
        // Add the widget to the widgets vector.
        this._widgets.pushBack(widget);
        // Insert the widget based on the specified mode.
        this._insertWidget(widget, mode, ref);
        // Activate the widget if requested.
        if (activate)
            this.activateWidget(widget);
    };
    /**
     * Find the drop target for the given client position.
     *
     * @param clientX - The client X position of interest.
     *
     * @param clientY - The client Y position of interest.
     *
     * @returns The drop target at the specified client position.
     */
    DockPanel.prototype.findDropTarget = function (clientX, clientY) {
        // If the position is not over the dock panel, bail.
        if (!query_1.hitTest(this.node, clientX, clientY)) {
            return { zone: 'invalid', panel: null };
        }
        // If there is no root, indicate a center drop position.
        if (!this._root) {
            return { zone: 'root-center', panel: null };
        }
        // Test for a root zone first.
        var zone = Private.getRootZone(this.node, clientX, clientY);
        if (zone !== 'invalid') {
            return { zone: zone, panel: null };
        }
        // Find the panel at the client position.
        var panel = searching_1.find(this._tabPanels, function (panel) {
            return query_1.hitTest(panel.node, clientX, clientY);
        }) || null;
        // Compute the zone for the hit panel, if any.
        if (panel) {
            zone = Private.getPanelZone(panel.node, clientX, clientY);
        }
        else {
            zone = 'invalid';
        }
        // Return the final drop target.
        return { zone: zone, panel: panel };
    };
    /**
     * Show the overlay indicator at the given client position.
     *
     * @param clientX - The client X position of interest.
     *
     * @param clientY - The client Y position of interest.
     *
     * @returns The dock zone at the specified client position.
     *
     * #### Notes
     * If the position is not over a valid zone, the overlay is hidden.
     *
     * This returns the dock zone used to display the overlay.
     */
    DockPanel.prototype.showOverlay = function (clientX, clientY) {
        // Find the dock target for the given client position.
        var target = this.findDropTarget(clientX, clientY);
        // If the dock zone is invalid, hide the overlay and bail.
        if (target.zone === 'invalid') {
            this._overlay.hide(100);
            return target.zone;
        }
        // Setup the variables needed to compute the overlay geometry.
        var top;
        var left;
        var right;
        var bottom;
        var cr;
        var box = sizing_1.boxSizing(this.node); // TODO cache this?
        var rect = this.node.getBoundingClientRect();
        // Compute the overlay geometry based on the dock zone.
        switch (target.zone) {
            case 'root-top':
                top = box.paddingTop;
                left = box.paddingLeft;
                right = box.paddingRight;
                bottom = (rect.height - box.verticalSum) * 2 / 3;
                break;
            case 'root-left':
                top = box.paddingTop;
                left = box.paddingLeft;
                right = (rect.width - box.horizontalSum) * 2 / 3;
                bottom = box.paddingBottom;
                break;
            case 'root-right':
                top = box.paddingTop;
                left = (rect.width - box.horizontalSum) * 2 / 3;
                right = box.paddingRight;
                bottom = box.paddingBottom;
                break;
            case 'root-bottom':
                top = (rect.height - box.verticalSum) * 2 / 3;
                left = box.paddingLeft;
                right = box.paddingRight;
                bottom = box.paddingBottom;
                break;
            case 'root-center':
                top = box.paddingTop;
                left = box.paddingLeft;
                right = box.paddingRight;
                bottom = box.paddingBottom;
                break;
            case 'panel-top':
                cr = target.panel.node.getBoundingClientRect();
                top = cr.top - rect.top - box.borderTop;
                left = cr.left - rect.left - box.borderLeft;
                right = rect.right - cr.right - box.borderRight;
                bottom = rect.bottom - cr.bottom + cr.height / 2 - box.borderBottom;
                break;
            case 'panel-left':
                cr = target.panel.node.getBoundingClientRect();
                top = cr.top - rect.top - box.borderTop;
                left = cr.left - rect.left - box.borderLeft;
                right = rect.right - cr.right + cr.width / 2 - box.borderRight;
                bottom = rect.bottom - cr.bottom - box.borderBottom;
                break;
            case 'panel-right':
                cr = target.panel.node.getBoundingClientRect();
                top = cr.top - rect.top - box.borderTop;
                left = cr.left - rect.left + cr.width / 2 - box.borderLeft;
                right = rect.right - cr.right - box.borderRight;
                bottom = rect.bottom - cr.bottom - box.borderBottom;
                break;
            case 'panel-bottom':
                cr = target.panel.node.getBoundingClientRect();
                top = cr.top - rect.top + cr.height / 2 - box.borderTop;
                left = cr.left - rect.left - box.borderLeft;
                right = rect.right - cr.right - box.borderRight;
                bottom = rect.bottom - cr.bottom - box.borderBottom;
                break;
            case 'panel-center':
                cr = target.panel.node.getBoundingClientRect();
                top = cr.top - rect.top - box.borderTop;
                left = cr.left - rect.left - box.borderLeft;
                right = rect.right - cr.right - box.borderRight;
                bottom = rect.bottom - cr.bottom - box.borderBottom;
                break;
        }
        // Derive the width and height from the other dimensions.
        var width = rect.width - right - left - box.borderLeft - box.borderRight;
        var height = rect.height - bottom - top - box.borderTop - box.borderBottom;
        // Show the overlay with the computed geometry.
        this._overlay.show({
            zone: target.zone,
            mouseX: clientX,
            mouseY: clientY,
            parentRect: rect,
            top: top, left: left, right: right, bottom: bottom, width: width, height: height
        });
        // Finally, return the dock zone used for the overlay.
        return target.zone;
    };
    /**
     * Handle the DOM events for the dock panel.
     *
     * @param event - The DOM event sent to the dock panel.
     *
     * #### Notes
     * This method implements the DOM `EventListener` interface and is
     * called in response to events on the dock panel's node. It should
     * not be called directly by user code.
     */
    DockPanel.prototype.handleEvent = function (event) {
        switch (event.type) {
            case 'p-dragenter':
                this._evtDragEnter(event);
                break;
            case 'p-dragleave':
                this._evtDragLeave(event);
                break;
            case 'p-dragover':
                this._evtDragOver(event);
                break;
            case 'p-drop':
                this._evtDrop(event);
                break;
        }
    };
    /**
     * A message handler invoked on an `'after-attach'` message.
     */
    DockPanel.prototype.onAfterAttach = function (msg) {
        var node = this.node;
        node.addEventListener('p-dragenter', this);
        node.addEventListener('p-dragleave', this);
        node.addEventListener('p-dragover', this);
        node.addEventListener('p-drop', this);
    };
    /**
     * A message handler invoked on a `'before-detach'` message.
     */
    DockPanel.prototype.onBeforeDetach = function (msg) {
        var node = this.node;
        node.removeEventListener('p-dragenter', this);
        node.removeEventListener('p-dragleave', this);
        node.removeEventListener('p-dragover', this);
        node.removeEventListener('p-drop', this);
    };
    /**
     * Handle the `'p-dragenter'` event for the dock panel.
     */
    DockPanel.prototype._evtDragEnter = function (event) {
        // If the factory mime type is present, mark the event as
        // handled in order to get the rest of the drag events.
        if (event.mimeData.hasData(FACTORY_MIME)) {
            event.preventDefault();
            event.stopPropagation();
        }
    };
    /**
     * Handle the `'p-dragleave'` event for the dock panel.
     */
    DockPanel.prototype._evtDragLeave = function (event) {
        // Always mark the event as handled.
        event.preventDefault();
        event.stopPropagation();
        // Get the node into which the drag is entering.
        var related = event.relatedTarget;
        // Hide the overlay if the drag is leaving the dock panel.
        if (!related || !this.node.contains(related)) {
            this._overlay.hide(0);
        }
    };
    /**
     * Handle the `'p-dragover'` event for the dock panel.
     */
    DockPanel.prototype._evtDragOver = function (event) {
        // Always mark the event as handled.
        event.preventDefault();
        event.stopPropagation();
        // Show the drop indicator overlay and update the drop
        // action based on the drop target zone under the mouse.
        if (this.showOverlay(event.clientX, event.clientY) === 'invalid') {
            event.dropAction = 'none';
        }
        else {
            event.dropAction = event.proposedAction;
        }
    };
    /**
     * Handle the `'p-drop'` event for the dock panel.
     */
    DockPanel.prototype._evtDrop = function (event) {
        // Always mark the event as handled.
        event.preventDefault();
        event.stopPropagation();
        // Hide the drop indicator overlay.
        this._overlay.hide(0);
        // Bail if the proposed action is to do nothing.
        if (event.proposedAction === 'none') {
            event.dropAction = 'none';
            return;
        }
        // Find the drop target under the mouse.
        var target = this.findDropTarget(event.clientX, event.clientY);
        // Bail if the drop zone is invalid.
        if (target.zone === 'invalid') {
            event.dropAction = 'none';
            return;
        }
        // Bail if the factory mime type has invalid data.
        var factory = event.mimeData.getData(FACTORY_MIME);
        if (typeof factory !== 'function') {
            event.dropAction = 'none';
            return;
        }
        // Bail if the factory does not produce a widget.
        var widget = factory();
        if (!(widget instanceof widget_1.Widget)) {
            event.dropAction = 'none';
            return;
        }
        // Handle the drop using the generated widget.
        this._handleDrop(widget, target);
        // Accept the proposed drop action.
        event.dropAction = event.proposedAction;
    };
    /**
     * Drop a widget onto the dock panel using the given drop target.
     */
    DockPanel.prototype._handleDrop = function (widget, target) {
        // Do nothing if the dock zone is invalid.
        if (target.zone === 'invalid') {
            return;
        }
        // Handle the simple case of root drops first.
        switch (target.zone) {
            case 'root-top':
                this.addWidget(widget, { mode: 'split-top' });
                return;
            case 'root-left':
                this.addWidget(widget, { mode: 'split-left' });
                return;
            case 'root-right':
                this.addWidget(widget, { mode: 'split-right' });
                return;
            case 'root-bottom':
                this.addWidget(widget, { mode: 'split-bottom' });
                return;
            case 'root-center':
                this.addWidget(widget, { mode: 'split-left' });
                return;
        }
        // Otherwise, it's a panel drop, and that requires more checks.
        // Fetch the children of the target panel.
        var children = target.panel.widgets;
        // Do nothing if the widget is dropped as a tab on its own panel.
        if (target.zone === 'panel-center' && searching_1.indexOf(children, widget) !== -1) {
            return;
        }
        // Do nothing if the panel only contains the drop widget.
        if (children.length === 1 && children.at(0) === widget) {
            return;
        }
        // Find a suitable reference widget for the drop.
        var ref = children.at(children.length - 1);
        if (ref === widget) {
            ref = children.at(children.length - 2);
        }
        // Insert the widget based on the panel zone.
        switch (target.zone) {
            case 'panel-top':
                this.addWidget(widget, { mode: 'split-top', ref: ref });
                return;
            case 'panel-left':
                this.addWidget(widget, { mode: 'split-left', ref: ref });
                return;
            case 'panel-right':
                this.addWidget(widget, { mode: 'split-right', ref: ref });
                return;
            case 'panel-bottom':
                this.addWidget(widget, { mode: 'split-bottom', ref: ref });
                return;
            case 'panel-center':
                this.addWidget(widget, { mode: 'tab-after', ref: ref });
                return;
        }
    };
    /**
     * Insert a widget into the dock panel using the given mode.
     *
     * The target widget should have no parent, and the reference widget
     * should either be null or a widget contained in the dock panel.
     */
    DockPanel.prototype._insertWidget = function (widget, mode, ref) {
        // Determine whether the insert is before or after the ref.
        var after = (mode === 'tab-after' ||
            mode === 'split-right' ||
            mode === 'split-bottom');
        // Handle the simple case of adding to a tab panel.
        if (mode === 'tab-before' || mode === 'tab-after') {
            if (ref) {
                var tabPanel_1 = ref.parent.parent;
                var index = searching_1.indexOf(tabPanel_1.widgets, ref) + (after ? 1 : 0);
                tabPanel_1.insertWidget(index, widget);
            }
            else {
                var tabPanel_2 = this._ensureTabPanel();
                var index = after ? tabPanel_2.widgets.length : 0;
                tabPanel_2.insertWidget(index, widget);
            }
            return;
        }
        // Otherwise, determine the orientation of the new split.
        var orientation;
        if (mode === 'split-top' || mode === 'split-bottom') {
            orientation = 'vertical';
        }
        else {
            orientation = 'horizontal';
        }
        // Setup the new tab panel to host the widget.
        var tabPanel = this._createTabPanel();
        tabPanel.addWidget(widget);
        // If there is no root, add the new tab panel as the root.
        if (!this._root) {
            this._setRoot(tabPanel);
            return;
        }
        // If the ref widget is null, split the root panel.
        if (!ref) {
            var splitPanel_1 = this._ensureSplitRoot(orientation);
            var sizes_1 = splitPanel_1.relativeSizes();
            var index = after ? sizes_1.length : 0;
            sizes_1.splice(index, 0, 0.5);
            splitPanel_1.insertWidget(index, tabPanel);
            splitPanel_1.setRelativeSizes(sizes_1);
            return;
        }
        // Lookup the tab panel for the ref widget.
        var refTabPanel = ref.parent.parent;
        // If the ref tab panel is the root, split the root.
        if (this._root === refTabPanel) {
            var splitPanel_2 = this._ensureSplitRoot(orientation);
            splitPanel_2.insertWidget(after ? 1 : 0, tabPanel);
            splitPanel_2.setRelativeSizes([1, 1]);
            return;
        }
        // Otherwise, the ref tab panel parent is a split panel.
        var splitPanel = refTabPanel.parent;
        // If the split panel is the correct orientation, the widget
        // can be inserted directly and sized to 0.5 of the ref space.
        if (splitPanel.orientation === orientation) {
            var i_1 = searching_1.indexOf(splitPanel.widgets, refTabPanel);
            var index = after ? i_1 + 1 : i_1;
            var sizes_2 = splitPanel.relativeSizes();
            var size = sizes_2[i_1] = sizes_2[i_1] / 2;
            sizes_2.splice(index, 0, size);
            splitPanel.insertWidget(index, tabPanel);
            splitPanel.setRelativeSizes(sizes_2);
            return;
        }
        // If the split panel only has a single child, its orientation
        // can be changed directly and its sizes set to a 1:1 ratio.
        if (splitPanel.widgets.length === 1) {
            splitPanel.orientation = orientation;
            splitPanel.insertWidget(after ? 1 : 0, tabPanel);
            splitPanel.setRelativeSizes([1, 1]);
            return;
        }
        // Otherwise, a new split panel with the correct orientation needs
        // to be created to hold the ref panel and tab panel, and inserted
        // at the previous location of the ref panel.
        var sizes = splitPanel.relativeSizes();
        var i = searching_1.indexOf(splitPanel.widgets, refTabPanel);
        var childSplit = this._createSplitPanel(orientation);
        childSplit.insertWidget(0, refTabPanel);
        childSplit.insertWidget(after ? 1 : 0, tabPanel);
        splitPanel.insertWidget(i, childSplit);
        splitPanel.setRelativeSizes(sizes);
        childSplit.setRelativeSizes([1, 1]);
    };
    /**
     * Create a new tab panel for adding to the dock panel.
     */
    DockPanel.prototype._createTabPanel = function () {
        var panel = new tabpanel_1.TabPanel({ tabsMovable: true });
        panel.addClass(TAB_PANEL_CLASS);
        panel.tabBar.tabDetachRequested.connect(this._onTabDetachRequested, this);
        panel.stackedPanel.widgetRemoved.connect(this._onWidgetRemoved, this);
        this._tabPanels.pushBack(panel);
        return panel;
    };
    /**
     * Create a new split panel for adding to the dock panel.
     */
    DockPanel.prototype._createSplitPanel = function (orientation) {
        var panel = new splitpanel_1.SplitPanel({ orientation: orientation, spacing: this._spacing });
        panel.addClass(SPLIT_PANEL_CLASS);
        this._splitPanels.pushBack(panel);
        return panel;
    };
    /**
     * Dispose of a tab panel created for the dock panel.
     */
    DockPanel.prototype._disposeTabPanel = function (panel) {
        this._tabPanels.remove(panel);
        panel.dispose();
    };
    /**
     * Dispose of a split panel created for the dock panel.
     */
    DockPanel.prototype._disposeSplitPanel = function (panel) {
        this._splitPanels.remove(panel);
        panel.dispose();
    };
    /**
     * Ensure a tab panel is available in the dock panel.
     *
     * If there is no root panel, a new root tab panel will be created.
     * Otherwise, if there is an active widget, the tab panel for that
     * widget will be returned. Otherwise, the top-left tab panel will
     * be returned.
     */
    DockPanel.prototype._ensureTabPanel = function () {
        // If there is no root panel, create a new root tab panel.
        if (!this._root) {
            var panel_1 = this._createTabPanel();
            this._setRoot(panel_1);
            return panel_1;
        }
        // Otherwise, use the tab panel of the current widget if possible.
        var current = this._tracker.currentWidget;
        if (current) {
            return current.parent.parent;
        }
        // Otherwise, fallback on using the top-left tab panel.
        var panel = this._root;
        while (panel instanceof splitpanel_1.SplitPanel) {
            panel = panel.widgets.at(0);
        }
        // Return the top-left tab panel.
        return panel;
    };
    /**
     * Ensure the root panel is a split panel with the given orientation.
     *
     * This returns split panel so that casting the root is not needed.
     */
    DockPanel.prototype._ensureSplitRoot = function (orientation) {
        // If there is no root panel, create a new root split panel.
        if (!this._root) {
            var root = this._createSplitPanel(orientation);
            this._setRoot(root);
            return root;
        }
        // If the root is a tab panel, add it to a new split root.
        if (this._root instanceof tabpanel_1.TabPanel) {
            var root = this._createSplitPanel(orientation);
            root.addWidget(this._root);
            this._setRoot(root);
            return root;
        }
        // Otherwise, do nothing if the split root orientation is okay.
        var oldRoot = this._root;
        if (oldRoot.orientation === orientation) {
            return oldRoot;
        }
        // Correct the orientation of the split root, if feasible.
        if (oldRoot.widgets.length <= 1) {
            oldRoot.orientation = orientation;
            return oldRoot;
        }
        // Otherwise, create a new split root.
        var newRoot = this._createSplitPanel(orientation);
        newRoot.addWidget(oldRoot);
        this._setRoot(newRoot);
        return newRoot;
    };
    /**
     * Set the non-null root widget for the dock panel.
     */
    DockPanel.prototype._setRoot = function (root) {
        this._root = root;
        this.layout.addWidget(root);
        root.show();
    };
    /**
     * Remove an empty dock tab panel from the hierarchy.
     *
     * This ensures that the hierarchy is kept consistent by merging an
     * ancestor split panel when it contains only a single child widget.
     */
    DockPanel.prototype._removeTabPanel = function (tabPanel) {
        // If the parent of the tab panel is the root, just remove it.
        if (this._root === tabPanel) {
            this._disposeTabPanel(tabPanel);
            this._root = null;
            return;
        }
        // Cast the tab panel parent to a split panel.
        // It's guaranteed to have at least 2 children.
        var splitPanel = tabPanel.parent;
        // Release the tab panel.
        this._disposeTabPanel(tabPanel);
        // Do nothing more if the split panel still has multiple children.
        if (splitPanel.widgets.length > 1) {
            return;
        }
        // Extract the remaining child from the split panel.
        var child = splitPanel.widgets.at(0);
        // If the split panel is the root, replace it.
        if (this._root === splitPanel) {
            this._setRoot(child);
            this._disposeSplitPanel(splitPanel);
            return;
        }
        // Cast the split panel parent to a split panel and lookup
        // the index of the split panel. The grand panel will have
        // at least 2 children.
        var grandPanel = splitPanel.parent;
        var index = searching_1.indexOf(grandPanel.widgets, splitPanel);
        // If the child is a tab panel, replace the split panel.
        if (child instanceof tabpanel_1.TabPanel) {
            var sizes = grandPanel.relativeSizes();
            splitPanel.parent = null;
            grandPanel.insertWidget(index, child);
            grandPanel.setRelativeSizes(sizes);
            this._disposeSplitPanel(splitPanel);
            return;
        }
        // Cast the child to a split panel.
        var childSplit = child;
        // Child splitters have an orthogonal orientation to their parent.
        // The grand children can now be merged with their grand parent.
        // Start by fetching the relevant current sizes.
        var childSizes = childSplit.relativeSizes();
        var grandSizes = grandPanel.relativeSizes();
        // Remove the split panel and store its share of the size.
        splitPanel.parent = null;
        var sizeShare = grandSizes.splice(index, 1)[0];
        // Merge the grand children and maintain their relative size.
        for (var i = 0; childSplit.widgets.length !== 0; ++i) {
            grandPanel.insertWidget(index + i, childSplit.widgets.at(0));
            grandSizes.splice(index + i, 0, sizeShare * childSizes[i]);
        }
        // Update the grand parent sizes.
        grandPanel.setRelativeSizes(grandSizes);
        // Dispose the removed split panel.
        this._disposeSplitPanel(splitPanel);
    };
    /**
     * Handle the `tabDetachRequested` signal from a tab bar.
     */
    DockPanel.prototype._onTabDetachRequested = function (sender, args) {
        var _this = this;
        // Do nothing if a drag is already in progress.
        if (this._drag) {
            return;
        }
        // Release the tab bar's hold on the mouse.
        sender.releaseMouse();
        // Extract the data from the args.
        var index = args.index, title = args.title, tab = args.tab, clientX = args.clientX, clientY = args.clientY;
        // Setup the mime data for the drag operation.
        var mimeData = new mimedata_1.MimeData();
        var widget = title.owner;
        mimeData.setData(FACTORY_MIME, function () { return widget; });
        // Create the drag image for the drag operation.
        var dragImage = tab.cloneNode(true);
        // Create the drag object to manage the drag-drop operation.
        this._drag = new dragdrop_1.Drag({
            mimeData: mimeData,
            dragImage: dragImage,
            proposedAction: 'move',
            supportedActions: 'move',
        });
        // Hide the tab node in the original tab.
        tab.classList.add(HIDDEN_CLASS);
        // Create the cleanup callback.
        var cleanup = (function () {
            _this._drag = null;
            tab.classList.remove(HIDDEN_CLASS);
        });
        // Start the drag operation and cleanup when done.
        this._drag.start(clientX, clientY).then(cleanup);
    };
    /**
     * Handle the `widgetRemoved` signal from a stacked panel.
     */
    DockPanel.prototype._onWidgetRemoved = function (sender, widget) {
        this._widgets.remove(widget);
        this._tracker.remove(widget);
        if (sender.widgets.length === 0) {
            this._removeTabPanel(sender.parent);
        }
    };
    /**
     * Handle the `currentChanged` signal from the focus tracker.
     */
    DockPanel.prototype._onCurrentChanged = function (sender, args) {
        this.currentChanged.emit(args);
    };
    return DockPanel;
}(widget_1.Widget));
exports.DockPanel = DockPanel;
// Define the signals for the `FocusTracker` class.
signaling_1.defineSignal(DockPanel.prototype, 'currentChanged');
/**
 * The namespace for the `DockPanel` class statics.
 */
var DockPanel;
(function (DockPanel) {
    /**
     * A concrete implementation of `IOverlay`.
     *
     * This is the default overlay implementation for a dock panel.
     */
    var Overlay = (function () {
        /**
         * Construct a new overlay.
         */
        function Overlay() {
            this._timer = -1;
            this._hidden = true;
            this._zone = 'invalid';
            this._node = document.createElement('div');
            this._node.classList.add(OVERLAY_CLASS);
            this._node.classList.add(HIDDEN_CLASS);
            this._node.style.position = 'absolute';
        }
        Object.defineProperty(Overlay.prototype, "node", {
            /**
             * The DOM node for the overlay.
             *
             * #### Notes
             * This is a read-only property.
             */
            get: function () {
                return this._node;
            },
            enumerable: true,
            configurable: true
        });
        /**
         * Show the overlay using the given overlay geometry.
         *
         * @param geo - The desired geometry for the overlay.
         */
        Overlay.prototype.show = function (geo) {
            // Update the position of the overlay.
            var style = this._node.style;
            style.top = geo.top + "px";
            style.left = geo.left + "px";
            style.right = geo.right + "px";
            style.bottom = geo.bottom + "px";
            // Update the class name for the zone.
            this._setZone(geo.zone);
            // Clear any pending hide timer.
            clearTimeout(this._timer);
            this._timer = -1;
            // If the overlay is already visible, we're done.
            if (!this._hidden) {
                return;
            }
            // Clear the hidden flag.
            this._hidden = false;
            // Finally, show the overlay.
            this._node.classList.remove(HIDDEN_CLASS);
        };
        /**
         * Hide the overlay node.
         *
         * @param delay - The delay (in ms) before hiding the overlay.
         *   A delay value <= 0 will hide the overlay immediately.
         */
        Overlay.prototype.hide = function (delay) {
            var _this = this;
            // Do nothing if the overlay is already hidden.
            if (this._hidden) {
                return;
            }
            // Hide immediately if the delay is <= 0.
            if (delay <= 0) {
                clearTimeout(this._timer);
                this._timer = -1;
                this._hidden = true;
                this._node.classList.add(HIDDEN_CLASS);
                return;
            }
            // Do nothing if a hide is already pending.
            if (this._timer !== -1) {
                return;
            }
            // Otherwise setup the hide timer.
            this._timer = setTimeout(function () {
                _this._timer = -1;
                _this._hidden = true;
                _this._node.classList.add(HIDDEN_CLASS);
            }, delay);
        };
        /**
         * Set the dock zone for the overlay.
         */
        Overlay.prototype._setZone = function (newZone) {
            var oldZone = this._zone;
            if (oldZone === newZone) {
                return;
            }
            var oldClass = oldZone !== 'invalid' ? "p-mod-" + oldZone : '';
            var newClass = newZone !== 'invalid' ? "p-mod-" + newZone : '';
            if (oldClass)
                this._node.classList.remove(oldClass);
            if (newClass)
                this._node.classList.add(newClass);
            this._zone = newZone;
        };
        return Overlay;
    }());
    DockPanel.Overlay = Overlay;
})(DockPanel = exports.DockPanel || (exports.DockPanel = {}));
/**
 * The namespace for the module private data.
 */
var Private;
(function (Private) {
    /**
     * Clamp a spacing value to an integer >= 0.
     */
    function clampSpacing(value) {
        return Math.max(0, Math.floor(value));
    }
    Private.clampSpacing = clampSpacing;
    /**
     * Get the root zone for the given node and client position.
     *
     * This assumes the position lies within the node's client rect.
     *
     * Returns the `'invalid'` zone if the position is not within an edge.
     */
    function getRootZone(node, x, y) {
        var zone;
        var rect = node.getBoundingClientRect();
        if (x < rect.left + EDGE_SIZE) {
            if (y - rect.top < x - rect.left) {
                zone = 'root-top';
            }
            else if (rect.bottom - y < x - rect.left) {
                zone = 'root-bottom';
            }
            else {
                zone = 'root-left';
            }
        }
        else if (x >= rect.right - EDGE_SIZE) {
            if (y - rect.top < rect.right - x) {
                zone = 'root-top';
            }
            else if (rect.bottom - y < rect.right - x) {
                zone = 'root-bottom';
            }
            else {
                zone = 'root-right';
            }
        }
        else if (y < rect.top + EDGE_SIZE) {
            zone = 'root-top';
        }
        else if (y >= rect.bottom - EDGE_SIZE) {
            zone = 'root-bottom';
        }
        else {
            zone = 'invalid';
        }
        return zone;
    }
    Private.getRootZone = getRootZone;
    /**
     * Get the panel zone for the given node and position.
     *
     * This assumes the position lies within the node's client rect.
     *
     * This always returns a valid zone.
     */
    function getPanelZone(node, x, y) {
        var zone;
        var rect = node.getBoundingClientRect();
        var fracX = (x - rect.left) / rect.width;
        var fracY = (y - rect.top) / rect.height;
        if (fracX < 1 / 3) {
            if (fracY < fracX) {
                zone = 'panel-top';
            }
            else if (1 - fracY < fracX) {
                zone = 'panel-bottom';
            }
            else {
                zone = 'panel-left';
            }
        }
        else if (fracX < 2 / 3) {
            if (fracY < 1 / 3) {
                zone = 'panel-top';
            }
            else if (fracY < 2 / 3) {
                zone = 'panel-center';
            }
            else {
                zone = 'panel-bottom';
            }
        }
        else {
            if (fracY < 1 - fracX) {
                zone = 'panel-top';
            }
            else if (fracY > fracX) {
                zone = 'panel-bottom';
            }
            else {
                zone = 'panel-right';
            }
        }
        return zone;
    }
    Private.getPanelZone = getPanelZone;
})(Private || (Private = {}));
